import matplotlib
import matplotlib.pyplot as plt
from numpy import argsort
from scipy.signal import find_peaks as peak_find

matplotlib.use('TkAgg')
import logging

from . import utilities as utili

def peak_finder(data, prominence, height=0, width=0):
    logging.debug(f'finding peaks based on prominence of {prominence}')
    # do a routine looping through until the right number of peaks is found

    peaks, properties = peak_find(data, prominence=prominence, height=height, width=width)  # find the peak positions in the data

    peaks = list(peaks)  # convert to a list
    amps = list(properties['peak_heights'])  # store the heights
    sorted_indices = argsort(amps)[::-1] # we will sort below in order of amplitudes

    peak_info = {'center_indices': sort_lists(sorted_indices, peaks), 'right_edges': sort_lists(sorted_indices, list(properties['right_bases'])),
                 'left_edges': sort_lists(sorted_indices, list(properties['left_bases'])), 'amplitude': sort_lists(sorted_indices, amps)}
    return peak_info

def sort_lists(sorted_indices, list_to_sort):
    return [list_to_sort[i] for i in sorted_indices]

def find_cents(prominence, y_data, find_all=False):
    logging.debug(f'finding centre values based on prominence of {prominence}')
    peak_info = peak_finder(y_data, prominence, height=0, width=0)  # find the peak centers
    if find_all:
        return peak_info
    #center_indices = peak_info['center_indices']
    return peak_info

def auto_peak_finder(prominence, x_data, y_data):
    while True:
        peak_info = find_cents(prominence, y_data, find_all=True)
        plt.plot(x_data, y_data)
        peak_coordinates = [x_data[ind] for ind in peak_info['center_indices']]
        for xc in peak_coordinates:
            plt.axvline(x=xc)
        print(f"Peak finder requires user input, please look at the following plot with prominence={prominence}")
        plt.show()
        ans = input("If you are happy with the plot, type y. If not then please type a new prominence ")
        if ans == 'y':
            break
        else:
            try:
                prominence = float(ans)
            except ValueError:
                print("You entered an incorrect answer! Trying again...")
    plt.close()
    return peak_info, prominence

def user_peak_finder(x_data, y_data):
    peak_info = {'center_values':(), 'amplitude':(), 'sigma':()}
    for key, peak_value in peak_info.items():
        while True:
            to_plot = peak_value
            if key == 'sigma':
                cents = peak_info['center_values']
                to_plot = [(cents[i] - width/2, cents[i] + width/2) for i, width in enumerate(to_plot)]
            print(f"Peak finder requires user input, please look at the following plot of values for {key}: {peak_value}")
            plot_values(key, to_plot, x_data, y_data)
            ans = input("If you are happy with the plot, type y. If not then please type a list of values seperated by commas ")
            if ans == 'y':
                break
            else:
                try:
                    peak_value_ = ans.split(',')
                    peak_value = [float(i) for i in peak_value_]
                except ValueError:
                    print("You entered an incorrect answer! Trying again...")
        plt.close()
        peak_info[key] = peak_value
    return peak_info

def plot_values(key, list_of_values, x_data, y_data):
    plt.plot(x_data, y_data)
    if key == 'center_values':
        for xc in list_of_values:
            plt.axvline(x=xc)
    elif key == 'amplitude':
        for yc in list_of_values:
            plt.axhline(y=yc)
    elif key =='sigma': # these have to be passed in as band edges
        for band in list_of_values:
            plt.axvspan(band[0], band[1], alpha=0.1, color='red')
    plt.show()

def interactive_peakfinder(x_data, y_data, type='auto', prominence=1):
    logging.debug('finding centres of peaks with user guided routine')
    if type == 'user':
        peak_info = user_peak_finder(x_data, y_data)
        prominence = None
    else:
        peak_info, prominence = auto_peak_finder(prominence, x_data, y_data)
        center_indices = peak_info['center_indices']
        peak_info['center_values'] = x_data[center_indices]
    return peak_info, prominence

def match_peak_centres(center_indices, y_data, prominence=1):
    while True:
        peak_info_ = find_cents(prominence, y_data, find_all=True)
        center_indices_ = utili.find_closest_indices(peak_info_['center_indices'],
                                                     center_indices)  # the indices might be slightly off so fix that
        if len(center_indices_) == len(center_indices):
            break
        elif len(center_indices_) > len(center_indices):
            center_indices_ = center_indices_[:len(center_indices)]  # as they're in order of prominence
            break
        else:
            prominence *= 10  # increase prominence until we get more than or equal to number.
        # do something smarter!
    center_indices = [peak_info_['center_indices'][i] for i in center_indices_]
    return center_indices

def find_peak_details(x_data, y_data, no_peaks, type='auto', prominence=1):
    logging.debug(f'finding peak details based on prominence of {prominence}, and user provided details:'
                  f'no_peaks:{no_peaks}')

    peak_info, prominence = interactive_peakfinder(x_data, y_data, type, prominence)
    no_peaks = max(len(value) for value in peak_info.values())
    center = peak_info['center_values']
    amplitude = peak_info['amplitude']
    sigma = peak_info.get('sigma', ())
    for i, sig in enumerate(sigma):
        x1 = x_data[0] + sig
        sigma[i] = utili.find_closest_indices(x_data, x1)
        import ipdb
        ipdb.set_trace()
    type = ["LorentzianModel" for i in range(no_peaks)]
    peak_info_dict = {}

    peak_info_dict['center'], peak_info_dict['amplitude'], peak_info_dict['sigma'], \
    peak_info_dict['type'] = center, amplitude, sigma, type

    return no_peaks, peak_info_dict, prominence

def make_bounds(x_data, y_data, no_peaks, peak_info_dict):
    logging.debug(f'making bounds based on: no_peaks:{no_peaks}, '
                  f'peak_info_dict: {peak_info_dict}')
    bounds = {}
    for key in peak_info_dict:
        while True:
            ans = input(f"Do you want to create bounds for {key}? \n"
                        f"Current values are: {peak_info_dict[key]} \n"
                        f"type y or n")
            if ans == 'y':
                bounds[key] = interactive_bounds(x_data, y_data, len(peak_info_dict[key]))
                break
            elif ans == 'n':
                break
            else:
                print("incorrect answer, please type y or n")
    return bounds

def interactive_bounds(x_data, y_data, no_bounds:int):
    bounds = [(0, 0) for i in range(no_bounds)]
    while True:
        plt.plot(x_data, y_data)
        for bound in bounds:
            plt.axvspan(bound[0], bound[1], alpha=0.1, color='red')
        print(f"current bounds are: {bounds}")
        plt.show()
        ans = input(f"Please enter a list of values for the bounds, comma delimited. You must enter {2*no_bounds}"
                    f" numbers. type y if you are happy with these bounds")
        if ans == 'y':
            break
        else:
            try:
                bounds = ans.split(',')
                bounds = [float(i) for i in bounds]
                if len(bounds) != 2 * no_bounds:
                    raise ValueError("incorrect no of bounds passed")
                else:
                    bounds = [(bounds[2*i], bounds[2*i+1]) for i in range(no_bounds)]
            except Exception as e:
                print("You entered an incorrect answer! Trying again...")
    plt.close()
    return bounds


import matplotlib.pyplot as plt
from os.path import basename
from ast import literal_eval
import numpy as np

from . import utilities
from . import parsing
from . import multi_obj
from . import thundobj
from . import peak_finding

def main():
    args = parsing.parse_user_args()

    arguments = parsing.using_user_args(args)

    if arguments['map']:
        try:  # a user can pass in a list of filenames or just one
            file_name = basename(literal_eval(arguments['datapath'])[0])
        except SyntaxError:  # assume its just a string and not a list passed
            file_name = None
            log_name = arguments['datapath']
            arguments['datapath'] = f"['{arguments['datapath']}',]"  # as this is what multiobj needs
        bag = multi_obj.main(arguments)  # create a Thunder object

        bag.choose_spectrum() # get the user to choose which spectrum to create params from
        thund = bag.thunder_bag[bag.first]
    else:
        thund = thundobj.main(arguments)  # create a Thunder object


    thund.y_data = utilities.sharpening_routine(thund.x_data, thund.y_data)

    peak_info, _ = peak_finding.interactive_peakfinder(thund.x_data, thund.y_data, type='user')
    utilities.save_fit_report(peak_info, './', 'generated_params.txt')




import logging

from lmfit import models
from numpy import nan

from . import utilities as utils

def get_tols(method, tol):
    if method in ['leastsq', 'least_squares']:
        tols = {'ftol': tol, 'xtol': tol}  # warning: need to experiment with these/allow control maybe?
    else:
        tols = {'ftol': tol}
    return tols

def prep_algo(method, tols):
    if method == 'basinhopping' or method == 'ampgo':
        print('Warning: This is a very slow but thorough algorithm')
    if method == 'differential_evolution':
        all_bounds = True
    if method in ['leastsq', 'least_squares', 'nelder', 'cobyla']:
        tols  = get_tols(method, tols)
    return tols

def fit_peaks(x_data, y_data, peak_info_dict, bounds, method='leastsq', tol=0.0000001):
    impl_methods = ['leastsq', 'least_squares', 'nelder', 'lbfgsb', 'powell', 'cg', 'cobyla', 'bfgsb',
                    'differential_evolution', 'basinhopping', 'ampgo']
    if not method in impl_methods:
        raise ValueError(f"The method supplied is not supported. Available methods: {impl_methods}")
    logging.debug(f'fitting peaks:  {peak_info_dict}')

    tols = prep_algo(method, tol)
    model_specs = build_specs(peak_info_dict, bounds)
    model, peak_params = generate_model(model_specs)

    if method in ['leastsq', 'least_squares', 'nelder', 'cobyla']:
        peaks = model.fit(y_data, peak_params, x=x_data, method=method, fit_kws=tols)
        if not peaks.success:
            print('peaks failed to fit, raising tolerance by one order magnitude and trying again')
            tols = {key:value*10 for key, value in tols.items()} # try raising the tolerance if it fails by one order
            peaks = model.fit(y_data, peak_params, x=x_data, method=method, fit_kws=tols)
    else:
        peaks = model.fit(y_data, peak_params, x=x_data, method=method)

    peak_params = peaks.best_values
    if not peaks.success:
        print('peaks failed to fit')
        peak_params = {key: nan for key in peak_params}

    return model_specs, model, peak_params, peaks

def safe_list_get (l, idx, default):
    """fetch items safely from a list, if it isn't long enough return a default value"""
    try:
        return l[idx]
    except IndexError:
        return default

##### Call this with correct arguments, should always call with all the arguments!
def build_specs(peak_info_dict: dict={}, bounds: dict={}):
    """Build a specs list, which has the peak specification details for each element corresponding to a peak to fit.
    Note that each element is a dictionary containing either None, or a value for that peak i.e. the parameter guess,
    or the bounds for that value. Only some will be valid for each peak type, which will be handled elsewhere"""
    logging.debug('building specs')
    type_ = peak_info_dict.get('type', ())
    if 'PolynomialModel' in type_ or 'SplitLorentzianModel' in type_ or 'ExponentialGaussianModel' in type_:
        raise NotImplementedError("No support for this model yet: polynomialmodel, ExponentialGaussianModel and splitlorentzianmodel")
    if 'StepModel' in type_ or 'RectangleModel' in type_:
        raise NotImplementedError('warning: no support for step models yet')
    if 'ExpressionModel' in type_:
        raise NotImplementedError("no support for expression models yet!")
    center = peak_info_dict.get('center', ())
    amplitude = peak_info_dict.get('amplitude', ())
    sigma = peak_info_dict.get('sigma', ())
    sigma_l = peak_info_dict.get('sigma_l', ())
    sigma_r = peak_info_dict.get('sigma_r', ())
    gamma = peak_info_dict.get('gamma', ())
    fraction = peak_info_dict.get('fraction', ())
    beta = peak_info_dict.get('beta', ())
    exponent = peak_info_dict.get('exponent', ())
    q = peak_info_dict.get('q', ())
    c = peak_info_dict.get('c', ())
    intercept = peak_info_dict.get('intercept', ())
    slope = peak_info_dict.get('slope', ())
    a = peak_info_dict.get('a', ())
    b = peak_info_dict.get('b', ())
    degree = peak_info_dict.get('degree', ())
    #poly_consts = peak_info_dict.get('poly_consts', ())
    form = peak_info_dict.get('form', ())
    center1 = peak_info_dict.get('center1', ())
    center2 = peak_info_dict.get('center2', ())
    sigma1 = peak_info_dict.get('sigma1', ())
    sigma2 = peak_info_dict.get('sigma2', ())
    decay = peak_info_dict.get('decay', ())
    expr = peak_info_dict.get('expr', ())

    specs = [
        {
            'type': type_[i],
            'params': {'center': utils.safe_list_get(center, i, None),
                       'amplitude': utils.safe_list_get(amplitude, i, None),
                       'sigma': utils.safe_list_get(sigma, i, None),
                       'sigma_r': utils.safe_list_get(sigma_r, i, None),
                       'sigma_l': utils.safe_list_get(sigma_l, i, None),
                       'gamma': utils.safe_list_get(gamma, i, None),
                       'fraction': utils.safe_list_get(fraction, i, None),
                       'beta': utils.safe_list_get(beta, i, None),
                       'exponent': utils.safe_list_get(exponent, i, None),
                       'q': utils.safe_list_get(q, i, None),
                       'c': utils.safe_list_get(c, i, None),
                       'intercept': utils.safe_list_get(intercept, i, None),
                       'slope': utils.safe_list_get(slope, i, None),
                       'a': utils.safe_list_get(a, i, None),
                       'b': utils.safe_list_get(b, i, None),
                       'degree': utils.safe_list_get(degree, i, None),

                       'form': utils.safe_list_get(form, i, None),
                       'center1': utils.safe_list_get(center1, i, None),
                       'center2': utils.safe_list_get(center2, i, None),
                       'sigma1': utils.safe_list_get(sigma1, i, None),
                       'sigma2': utils.safe_list_get(sigma2, i, None),
                       'decay': utils.safe_list_get(decay, i, None),
                       },
            'bounds': {'center': utils.safe_list_get(bounds.get('center', []), i,(None, None)),
                       'amplitude': utils.safe_list_get(bounds.get('amplitude', []), i, (None, None)),
                       'sigma': utils.safe_list_get(bounds.get('sigma', []), i, (None, None)),
                       'sigma_r': utils.safe_list_get(bounds.get('sigma_r', []), i, (None, None)),
                       'gamma': utils.safe_list_get(bounds.get('gamma', []), i, (None, None)),
                       'fraction': utils.safe_list_get(bounds.get('fraction', []), i, (None, None)),
                       'beta': utils.safe_list_get(bounds.get('beta', []), i, (None, None)),
                       'exponent': utils.safe_list_get(bounds.get('exponent', []), i, (None, None)),
                       'q': utils.safe_list_get(bounds.get('q', []), i, (None, None)),
                       'c': utils.safe_list_get(bounds.get('c', []), i, (None, None)),
                       'intercept': utils.safe_list_get(bounds.get('intercept', []), i, (None, None)),
                       'slope': utils.safe_list_get(bounds.get('slope', []), i, (None, None)),
                       'a': utils.safe_list_get(bounds.get('a', []), i, (None, None)),
                       'b': utils.safe_list_get(bounds.get('b', []), i, (None, None)),
                       'degree': utils.safe_list_get(bounds.get('degree', []), i, (None, None)),

                       'form': utils.safe_list_get(bounds.get('form', []), i, (None, None)),
                       'center1': utils.safe_list_get(bounds.get('center1', []), i, (None, None)),
                       'center2': utils.safe_list_get(bounds.get('center2', []), i, (None, None)),
                       'sigma1': utils.safe_list_get(bounds.get('sigma1', []), i, (None, None)),
                       'sigma2': utils.safe_list_get(bounds.get('sigma2', []), i, (None, None)),
                       'decay': utils.safe_list_get(bounds.get('decay', []), i, (None, None))
                       },
            'expr': utils.safe_list_get(expr, i, None)
        }
        for i in range(len(type_))
    ]



    #'poly_consts': utils.safe_list_get(poly_consts, i, None), and
    # 'poly_consts': utils.safe_list_get(bounds.get('poly_consts', []), i, [None, None]),
    # not yet implemented!

    return specs

def generate_model(model_specs):
    """
    https://chrisostrouchov.com/post/peak_fit_xrd_python/
    :param model_specs:
    :return:
    """
    logging.debug('generating model specs')
    composite_model = None
    params = None
    for i, spec in enumerate(model_specs):
        prefix = f'm{i}__'

        if spec['type'] == 'ExpressionModel':
            expr = spec['expr']
            model = models.ExpressionModel(expr)
        else:
            model = getattr(models, spec['type'])(prefix=prefix)  # generate the lmfit model based on the type specified
        model = decide_model_actions(spec, model, prefix) # call another function to decide what to do
        model_params = model.make_params() # make the params object
        if params is None: # first loop
            params = model_params
            composite_model = model
        else: # subsequent loops
            params.update(model_params)
            composite_model = composite_model + model
    return composite_model, params

def decide_model_actions(spec, model, prefix):
    for param_key, param_value in spec['params'].items():
        if param_value: # then set this value
            p_name = prefix + param_key
            model.set_param_hint(param_key, value=param_value)
    for bound_key, bound_value in spec['bounds'].items():
        if bound_value[0]: # then set lower bound
            b_name = prefix + bound_key
            model.set_param_hint(bound_key, min=bound_value[0])
        if bound_value[1]: # then set upper bound
            b_name = prefix + bound_key
            model.set_param_hint(bound_key, max=bound_value[1])
    return model




import logging
from copy import deepcopy
from difflib import get_close_matches
from re import findall
from typing import Dict, Union

import matplotlib.pyplot as plt
from lmfit.model import ModelResult
from numpy import ndarray

from . import plotting
from . import utilities as utili
from .background import background_removal as bg_remove
from . import peak_finding
from . import peak_fitting

# TODO: need to fail if peak fitting doesn't work!

class Thunder():
    """
    thunder object with all the methods we love inside it. Name generated using WuTang Clan name generator.
    """

    def __init__(self, input, x_data=None, y_data=None, e_data=None):
        self.input: Union[Thunder, Dict] = input

        self.x_ind: int = 0
        self.y_ind: int = 1
        self.e_ind: Union[int, None] = None

        self.x_data: Union[None, ndarray] = x_data
        self.y_data: Union[None, ndarray] = y_data
        self.e_data: Union[None, ndarray] = e_data

        self.y_data_bg_rm = None
        self.y_data_norm = None

        self.datapath: str = './data.txt'

        self.no_peaks: int = 0
        self.background: str = "SCARF"  # fix this
        self.scarf_params: Union[None, Dict] = None

        self.peak_info_dict: Dict = {}
        self.peak_finder_type: str = 'auto'

        self.bounds: Union[None, Dict] = None

        self.peaks: ModelResult
        self.plot: plt = None
        self.fit_report: {} = {}
        self.peak_params: {} = {}
        self.model = None  # give type!

        self.free_params: int = 0
        self.p_value: int = 0
        self.chisq: int = 0

        self.method: str = 'leastsq'
        self.tol: float = 0.0000001

        if isinstance(input, Thunder):  # if only pass one but its already a thunder object then just use that
            self.overwrite_thunder(input)  # add all the details in depending on args
        elif isinstance(input, dict):
            self.create_thunder(input)  # add all the details in depending on args
        else:
            raise TypeError('Cannot convert input to Thunder object')

        if isinstance(self.x_data, ndarray) and isinstance(self.y_data, ndarray):
            pass  # they're already loaded as they've been passed
        else:
            self.x_data, self.y_data, self.e_data = utili.load_data(self.datapath, self.x_ind,
                                                                    self.y_ind)  # load the data

    def overwrite_thunder(self, inp):
        logging.debug('overwriting thund obj')
        thun = inp

        if thun.x_data and thun.y_data:
            self.x_data = thun.x_data
            self.y_data = thun.y_data
        else:
            self.x_ind = thun.x_ind
            self.y_ind = thun.y_ind
            self.e_ind = thun.e_ind
            self.datapath = thun.datapath

        if thun.y_data_bg_rm:
            self.y_data_bg_rm = thun.y_data_bg_rm
        if thun.y_data_norm:
            self.y_data_norm = thun.y_data_norm

        self.no_peaks = thun.no_peaks
        self.background = thun.background
        self.scarf_params = thun.scarf_params

        self.peak_info_dict = thun.peak_info_dict
        self.peak_finder_type = thun.peak_finder_type

        self.bounds = thun.bounds

        self.method = thun.method
        self.tol = thun.tol

    def create_thunder(self, inp: Dict):
        """
        Used to create a thunder object given different input types
        :param args: a,b,c depending on type of input and
        :return: None, we modify the object unless a spec1d object is passed, in which case we return that
        """
        logging.debug('creating thund obj')
        try:  # only continue if its e_ind missing
            self.e_ind = inp['e_ind']
        except KeyError as e:
            logging.info(f"KeyError: Missing field in the data dictionary: {e}")

        try:  # If its the others you need to fail here
            self.datapath = inp['datapath']
            self.x_ind = inp['x_ind']
            self.y_ind = inp['y_ind']
        except KeyError as e:
            raise KeyError(f"Missing vital information to load object: {e}")

        self.no_peaks = inp.get('no_peaks', self.no_peaks)
        self.background = inp.get('background', self.background)
        # do some check on background here to set it to an np array
        self.scarf_params = inp.get('scarf_params', self.scarf_params)

        self.peak_info_dict = inp.get('peak_info_dict', self.peak_info_dict)
        self.peak_finder_type = inp.get('peak_finder_type', self.peak_finder_type)

        self.bounds = inp.get('bounds', self.bounds)
        self.method = inp.get('method', self.method)
        self.tol = inp.get('tol', self.tol)

    def clip_data(self):
        clip_left, clip_right = utili.clip_data(getattr(self, 'x_data'), getattr(self, 'y_data'))
        setattr(self, 'x_data', getattr(self, 'x_data')[clip_left:clip_right])
        setattr(self, 'y_data', getattr(self, 'y_data')[clip_left:clip_right])

    def remove_bg(self):
        background, y_data_bg_rm, params = bg_remove.background_finder(getattr(self, 'x_data'),
                                                                       getattr(self, 'y_data'),
                                                                       getattr(self, 'background'),
                                                                       getattr(self, 'scarf_params'))
        setattr(self, 'background', background)
        setattr(self, 'y_data_bg_rm', y_data_bg_rm)
        setattr(self, 'params', params)

    def normalise(self):
        y_data_bg_rm, background, y_data_norm = utili.normalise_all('y_data_bg_rm', 'background', 'y_data')
        setattr(self, 'background', background)
        setattr(self, 'y_data_bg_rm', y_data_bg_rm)
        setattr(self, 'y_data_norm', y_data_norm)

    def find_peaks(self):
        no_peaks, peak_info_dict, prominence = \
            peak_finding.find_peak_details(getattr(self, 'x_data'), getattr(self, 'y_data_bg_rm'),
                                           getattr(self, 'no_peaks'),
                                           getattr(self, 'peak_finder_type', 'auto'))
        setattr(self, 'no_peaks', no_peaks)  # set values
        center_indices = utili.find_closest_indices(self.x_data, peak_info_dict['center'])  # get the indices from the x centres
        center_indices = peak_finding.match_peak_centres(center_indices, self.y_data)  # match to the peakfinding cents
        peak_centres = self.x_data[center_indices]  # convert back to x values
        peak_info_dict['center'] = peak_centres  # set values
        setattr(self, peak_info_dict, peak_info_dict)

    def bound_setter(self, bounds=None):
        if not bounds:
            bounds = peak_finding.make_bounds(getattr(self, 'x_data'), getattr(self, 'y_data'), getattr(self, 'no_peaks'),
                                              self.peak_info_dict)
        setattr(self, 'bounds', bounds)  # set values

    def fit_peaks(self):
        specs, model, peak_params, peaks = peak_fitting.fit_peaks(getattr(self, 'x_data'),
                                                                          getattr(self, 'y_data_bg_rm'),
                                                                          getattr(self, 'peak_info_dict'),
                                                                          getattr(self, 'bounds'),
                                                                          getattr(self, 'method'),
                                                                          getattr(self, 'tol'))
        setattr(self, 'specs', specs)
        setattr(self, 'model', model)
        setattr(self, 'peak_params', peak_params)
        setattr(self, 'peaks', peaks)

    ## plot_all and fit_report need imporovements e.g. to check which attributes exists in the object
    def plot_all(self):
        logging.debug('plotting all for thund obj')
        ax, plt = plotting.plot_fits(self.x_data, self.peaks.eval_components())  # plot each component of the model
        ax, plt = plotting.plot_background(self.x_data, self.background, ax)  # plot the background supplied by user
        ax, plt = plotting.plot_fit_sum(self.x_data, self.peaks.best_fit, self.background, ax)  # plot the fitted data
        try:
            ax, plt = plotting.plot_uncertainty_curve(self.x_data, self.peaks.eval_uncertainty(sigma=3),
                                                      self.peaks.best_fit, ax)  # plot a band of uncertainty
        except TypeError:
            logging.warning('There are not uncertainties available for some reason - '
                            'try lowering the tightness of automatic bounds')
        ax, plt = plotting.plot_data(self.x_data, self.y_data, ax)  # plot the raw data

        ax.minorticks_on()
        ax.grid(which='minor', alpha=0.2)
        ax.grid(which='major', alpha=0.5)

        self.plot = plt

    def gen_fit_report(self):
        logging.debug('genertaing fit report for thund obj')
        self.fit_report = {mod_no: {} for mod_no in range(len(self.peak_info_dict['peak_types']))}

        ## total fit data
        self.fit_report['chi_sq'] = self.chi_sq
        self.fit_report['free_params'] = self.free_params
        self.fit_report['p_value'] = 'not implemented'

        ## individual parameter data
        param_info = {"center": "center", "amplitude": "amplitude", "sigma": "sigma", "fwhm": False, "height": False}
        for parameter, param_obj in self.peaks.params.items():
            model_no = int(findall(r'\d+', parameter)[0])
            param_type = param_info[get_close_matches(parameter, param_info.keys())[0]]

            if param_type:
                value = param_obj.value
                err = param_obj.stderr
                type = self.peak_info_dict['peak_types'][model_no]
                bounds = self.bounds[param_type][model_no]

                fit_info = {"value": value,
                            "stderr": err,
                            "peak_type": type,
                            "bounds": bounds}

                self.fit_report[model_no][param_type] = fit_info


def main(arguments):
    thunder = Thunder(deepcopy(arguments))  # load object
    return thunder
